@@ -13,15 +13,20 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-
-import six.moves
 import time
 import unittest
 
+import six.moves
+
 import tenacity
-from tenacity import retry
 from tenacity import RetryError
 from tenacity import Retrying
+from tenacity import retry
+
+
+class TestBase(unittest.TestCase):
+    def test_repr(self):
+        repr(tenacity.BaseRetrying())
 
 
 class TestStopConditions(unittest.TestCase):
@@ -30,6 +35,46 @@ class TestStopConditions(unittest.TestCase):
         r = Retrying()
         self.assertFalse(r.stop(3, 6546))
 
+    def test_stop_any(self):
+        s = tenacity.stop_any(
+            tenacity.stop_after_delay(1),
+            tenacity.stop_after_attempt(4))
+        self.assertFalse(s(1, 0.1))
+        self.assertFalse(s(2, 0.2))
+        self.assertFalse(s(2, 0.8))
+        self.assertTrue(s(4, 0.8))
+        self.assertTrue(s(3, 1.8))
+        self.assertTrue(s(4, 1.8))
+
+    def test_stop_all(self):
+        s = tenacity.stop_all(
+            tenacity.stop_after_delay(1),
+            tenacity.stop_after_attempt(4))
+        self.assertFalse(s(1, 0.1))
+        self.assertFalse(s(2, 0.2))
+        self.assertFalse(s(2, 0.8))
+        self.assertFalse(s(4, 0.8))
+        self.assertFalse(s(3, 1.8))
+        self.assertTrue(s(4, 1.8))
+
+    def test_stop_or(self):
+        s = tenacity.stop_after_delay(1) | tenacity.stop_after_attempt(4)
+        self.assertFalse(s(1, 0.1))
+        self.assertFalse(s(2, 0.2))
+        self.assertFalse(s(2, 0.8))
+        self.assertTrue(s(4, 0.8))
+        self.assertTrue(s(3, 1.8))
+        self.assertTrue(s(4, 1.8))
+
+    def test_stop_and(self):
+        s = tenacity.stop_after_delay(1) & tenacity.stop_after_attempt(4)
+        self.assertFalse(s(1, 0.1))
+        self.assertFalse(s(2, 0.2))
+        self.assertFalse(s(2, 0.8))
+        self.assertFalse(s(4, 0.8))
+        self.assertFalse(s(3, 1.8))
+        self.assertTrue(s(4, 1.8))
+
     def test_stop_after_attempt(self):
         r = Retrying(stop=tenacity.stop_after_attempt(3))
         self.assertFalse(r.stop(2, 6546))
@@ -130,20 +175,32 @@ class TestWaitConditions(unittest.TestCase):
     def test_legacy_explicit_wait_type(self):
         Retrying(wait="exponential_sleep")
 
+    def test_wait_func_result(self):
+        captures = []
+
+        def wait_capture(attempt, delay, last_result=None):
+            captures.append(last_result)
+            return 1
+
+        def dying():
+            raise Exception("Broken")
+
+        r_attempts = 10
+        r = Retrying(wait=wait_capture, sleep=lambda secs: None,
+                     stop=tenacity.stop_after_attempt(r_attempts),
+                     reraise=True)
+        self.assertRaises(Exception, r.call, dying)
+        self.assertEqual(r_attempts - 1, len(captures))
+        self.assertTrue(all([r.failed for r in captures]))
+
     def test_wait_func(self):
         r = Retrying(wait=lambda attempt, delay: attempt * delay)
         self.assertEqual(r.wait(1, 5), 5)
         self.assertEqual(r.wait(2, 11), 22)
         self.assertEqual(r.wait(10, 100), 1000)
 
-    def test_wait_jitter(self):
-        r = Retrying(wait=tenacity.wait_jitter(3))
-        # Test it a few time since it's random
-        for i in six.moves.range(1000):
-            self.assertLess(r.wait(1, 5), 3)
-
     def test_wait_combine(self):
-        r = Retrying(wait=tenacity.wait_combine(tenacity.wait_jitter(3),
+        r = Retrying(wait=tenacity.wait_combine(tenacity.wait_random(0, 3),
                                                 tenacity.wait_fixed(5)))
         # Test it a few time since it's random
         for i in six.moves.range(1000):
@@ -152,7 +209,7 @@ class TestWaitConditions(unittest.TestCase):
             self.assertGreaterEqual(w, 5)
 
     def test_wait_double_sum(self):
-        r = Retrying(wait=tenacity.wait_jitter(3) + tenacity.wait_fixed(5))
+        r = Retrying(wait=tenacity.wait_random(0, 3) + tenacity.wait_fixed(5))
         # Test it a few time since it's random
         for i in six.moves.range(1000):
             w = r.wait(1, 5)
@@ -160,7 +217,7 @@ class TestWaitConditions(unittest.TestCase):
             self.assertGreaterEqual(w, 5)
 
     def test_wait_triple_sum(self):
-        r = Retrying(wait=tenacity.wait_fixed(1) + tenacity.wait_jitter(3) +
+        r = Retrying(wait=tenacity.wait_fixed(1) + tenacity.wait_random(0, 3) +
                      tenacity.wait_fixed(5))
         # Test it a few time since it's random
         for i in six.moves.range(1000):
@@ -169,17 +226,23 @@ class TestWaitConditions(unittest.TestCase):
             self.assertGreaterEqual(w, 6)
 
     def test_wait_arbitrary_sum(self):
-        r = Retrying(wait=sum([tenacity.wait_fixed(1), tenacity.wait_jitter(3),
-                               tenacity.wait_fixed(5), tenacity.wait_none()]))
+        r = Retrying(wait=sum([tenacity.wait_fixed(1),
+                               tenacity.wait_random(0, 3),
+                               tenacity.wait_fixed(5),
+                               tenacity.wait_none()]))
         # Test it a few time since it's random
         for i in six.moves.range(1000):
             w = r.wait(1, 5)
             self.assertLess(w, 9)
             self.assertGreaterEqual(w, 6)
 
-    def _assert_range(self, wait, min, max):
-        self.assertLess(wait, max)
-        self.assertGreaterEqual(wait, min)
+    def _assert_range(self, wait, min_, max_):
+        self.assertLess(wait, max_)
+        self.assertGreaterEqual(wait, min_)
+
+    def _assert_inclusive_range(self, wait, low, high):
+        self.assertLessEqual(wait, high)
+        self.assertGreaterEqual(wait, low)
 
     def test_wait_chain(self):
         r = Retrying(wait=tenacity.wait_chain(
@@ -196,9 +259,65 @@ class TestWaitConditions(unittest.TestCase):
             else:
                 self._assert_range(w, 8, 9)
 
+    def test_wait_random_exponential(self):
+        fn = tenacity.wait_random_exponential(0.5, 60.0)
+
+        for _ in six.moves.range(1000):
+            self._assert_inclusive_range(fn(0, 0), 0, 0.5)
+            self._assert_inclusive_range(fn(1, 0), 0, 1.0)
+            self._assert_inclusive_range(fn(2, 0), 0, 2.0)
+            self._assert_inclusive_range(fn(3, 0), 0, 4.0)
+            self._assert_inclusive_range(fn(4, 0), 0, 8.0)
+            self._assert_inclusive_range(fn(5, 0), 0, 16.0)
+            self._assert_inclusive_range(fn(6, 0), 0, 32.0)
+            self._assert_inclusive_range(fn(7, 0), 0, 60.0)
+            self._assert_inclusive_range(fn(8, 0), 0, 60.0)
+            self._assert_inclusive_range(fn(9, 0), 0, 60.0)
+
+        fn = tenacity.wait_random_exponential(10, 5)
+        for _ in six.moves.range(1000):
+            self._assert_inclusive_range(fn(0, 0), 0.00, 5.00)
+
+        # Default arguments exist
+        fn = tenacity.wait_random_exponential()
+        fn(0, 0)
+
+    def test_wait_random_exponential_statistically(self):
+        fn = tenacity.wait_random_exponential(0.5, 60.0)
+
+        attempt = []
+        for i in six.moves.range(10):
+            attempt.append(
+                [fn(i, 0) for _ in six.moves.range(4000)]
+            )
+
+        def mean(lst):
+            return float(sum(lst)) / float(len(lst))
+
+        self._assert_inclusive_range(mean(attempt[0]), 0.20, 0.30)
+        self._assert_inclusive_range(mean(attempt[1]), 0.35, 0.65)
+        self._assert_inclusive_range(mean(attempt[2]), 0.75, 1.25)
+        self._assert_inclusive_range(mean(attempt[3]), 1.75, 3.25)
+        self._assert_inclusive_range(mean(attempt[4]), 3.50, 5.50)
+        self._assert_inclusive_range(mean(attempt[5]), 7.00, 9.00)
+        self._assert_inclusive_range(mean(attempt[6]), 14.00, 18.00)
+        self._assert_inclusive_range(mean(attempt[7]), 28.00, 34.00)
+        self._assert_inclusive_range(mean(attempt[8]), 28.00, 34.00)
+        self._assert_inclusive_range(mean(attempt[9]), 28.00, 34.00)
+
 
 class TestRetryConditions(unittest.TestCase):
 
+    def test_retry_if_result(self):
+        r = (tenacity.retry_if_result(lambda x: x == 1))
+        self.assertTrue(r(tenacity.Future.construct(1, 1, False)))
+        self.assertFalse(r(tenacity.Future.construct(1, 2, False)))
+
+    def test_retry_if_not_result(self):
+        r = (tenacity.retry_if_not_result(lambda x: x == 1))
+        self.assertTrue(r(tenacity.Future.construct(1, 2, False)))
+        self.assertFalse(r(tenacity.Future.construct(1, 1, False)))
+
     def test_retry_any(self):
         r = tenacity.retry_any(
             tenacity.retry_if_result(lambda x: x == 1),
@@ -208,6 +327,31 @@ class TestRetryConditions(unittest.TestCase):
         self.assertFalse(r(tenacity.Future.construct(1, 3, False)))
         self.assertFalse(r(tenacity.Future.construct(1, 1, True)))
 
+    def test_retry_all(self):
+        r = tenacity.retry_all(
+            tenacity.retry_if_result(lambda x: x == 1),
+            tenacity.retry_if_result(lambda x: isinstance(x, int)))
+        self.assertTrue(r(tenacity.Future.construct(1, 1, False)))
+        self.assertFalse(r(tenacity.Future.construct(1, 2, False)))
+        self.assertFalse(r(tenacity.Future.construct(1, 3, False)))
+        self.assertFalse(r(tenacity.Future.construct(1, 1, True)))
+
+    def test_retry_and(self):
+        r = (tenacity.retry_if_result(lambda x: x == 1) &
+             tenacity.retry_if_result(lambda x: isinstance(x, int)))
+        self.assertTrue(r(tenacity.Future.construct(1, 1, False)))
+        self.assertFalse(r(tenacity.Future.construct(1, 2, False)))
+        self.assertFalse(r(tenacity.Future.construct(1, 3, False)))
+        self.assertFalse(r(tenacity.Future.construct(1, 1, True)))
+
+    def test_retry_or(self):
+        r = (tenacity.retry_if_result(lambda x: x == "foo") |
+             tenacity.retry_if_result(lambda x: isinstance(x, int)))
+        self.assertTrue(r(tenacity.Future.construct(1, "foo", False)))
+        self.assertFalse(r(tenacity.Future.construct(1, "foobar", False)))
+        self.assertFalse(r(tenacity.Future.construct(1, 2.2, False)))
+        self.assertFalse(r(tenacity.Future.construct(1, 42, True)))
+
     def _raise_try_again(self):
         self._attempts += 1
         if self._attempts < 3:
@@ -219,9 +363,32 @@ class TestRetryConditions(unittest.TestCase):
                  retry=tenacity.retry_never).call(self._raise_try_again)
         self.assertEqual(3, self._attempts)
 
+    def test_retry_try_again_forever(self):
+        def _r():
+            raise tenacity.TryAgain
+
+        r = Retrying(stop=tenacity.stop_after_attempt(5),
+                     retry=tenacity.retry_never)
+        self.assertRaises(tenacity.RetryError,
+                          r.call,
+                          _r)
+        self.assertEqual(5, r.statistics['attempt_number'])
+
+    def test_retry_try_again_forever_reraise(self):
+        def _r():
+            raise tenacity.TryAgain
+
+        r = Retrying(stop=tenacity.stop_after_attempt(5),
+                     retry=tenacity.retry_never,
+                     reraise=True)
+        self.assertRaises(tenacity.TryAgain,
+                          r,
+                          _r)
+        self.assertEqual(5, r.statistics['attempt_number'])
+
 
 class NoneReturnUntilAfterCount(object):
-    "Holds counter state for invoking a method several times in a row."
+    """Holds counter state for invoking a method several times in a row."""
 
     def __init__(self, count):
         self.counter = 0
@@ -239,7 +406,7 @@ class NoneReturnUntilAfterCount(object):
 
 
 class NoIOErrorAfterCount(object):
-    "Holds counter state for invoking a method several times in a row."
+    """Holds counter state for invoking a method several times in a row."""
 
     def __init__(self, count):
         self.counter = 0
@@ -257,7 +424,7 @@ class NoIOErrorAfterCount(object):
 
 
 class NoNameErrorAfterCount(object):
-    "Holds counter state for invoking a method several times in a row."
+    """Holds counter state for invoking a method several times in a row."""
 
     def __init__(self, count):
         self.counter = 0
@@ -266,7 +433,7 @@ class NoNameErrorAfterCount(object):
     def go(self):
         """Raise a NameError until after count threshold has been crossed.
 
-        Tthen return True.
+        Then return True.
         """
         if self.counter < self.count:
             self.counter += 1
@@ -274,6 +441,42 @@ class NoNameErrorAfterCount(object):
         return True
 
 
+class NameErrorUntilCount(object):
+    """Holds counter state for invoking a method several times in a row."""
+
+    def __init__(self, count):
+        self.counter = 0
+        self.count = count
+
+    def go(self):
+        """Return True until after count threshold has been crossed.
+
+        Then raise a NameError.
+        """
+        if self.counter < self.count:
+            self.counter += 1
+            return True
+        raise NameError("Hi there, I'm a NameError")
+
+
+class IOErrorUntilCount(object):
+    """Holds counter state for invoking a method several times in a row."""
+
+    def __init__(self, count):
+        self.counter = 0
+        self.count = count
+
+    def go(self):
+        """Return True until after count threshold has been crossed.
+
+        Then raise an IOError.
+        """
+        if self.counter < self.count:
+            self.counter += 1
+            return True
+        raise IOError("Hi there, I'm an IOError")
+
+
 class CustomError(Exception):
     """This is a custom exception class.
 
@@ -293,14 +496,14 @@ class CustomError(Exception):
 
 
 class NoCustomErrorAfterCount(object):
-    "Holds counter state for invoking a method several times in a row."
+    """Holds counter state for invoking a method several times in a row."""
 
     def __init__(self, count):
         self.counter = 0
         self.count = count
 
     def go(self):
-        """Raise a CustomError until after count threshold has been crossed
+        """Raise a CustomError until after count threshold has been crossed.
 
         Then return True.
         """
@@ -339,6 +542,23 @@ def _retryable_test_with_exception_type_io_attempt_limit(thing):
     return thing.go()
 
 
+@retry(retry=tenacity.retry_unless_exception_type(NameError))
+def _retryable_test_with_unless_exception_type_name(thing):
+    return thing.go()
+
+
+@retry(
+    stop=tenacity.stop_after_attempt(3),
+    retry=tenacity.retry_unless_exception_type(NameError))
+def _retryable_test_with_unless_exception_type_name_attempt_limit(thing):
+    return thing.go()
+
+
+@retry(retry=tenacity.retry_unless_exception_type())
+def _retryable_test_with_unless_exception_type_no_input(thing):
+    return thing.go()
+
+
 @retry
 def _retryable_default(thing):
     return thing.go()
@@ -370,6 +590,14 @@ class TestDecoratorWrapper(unittest.TestCase):
         self.assertGreaterEqual(t, 250)
         self.assertTrue(result)
 
+    def test_retry_with(self):
+        start = current_time_ms()
+        result = _retryable_test_with_wait.retry_with(
+            wait=tenacity.wait_fixed(0.1))(NoneReturnUntilAfterCount(5))
+        t = current_time_ms() - start
+        self.assertGreaterEqual(t, 500)
+        self.assertTrue(result)
+
     def test_with_stop_on_return_value(self):
         try:
             _retryable_test_with_stop(NoneReturnUntilAfterCount(5))
@@ -410,6 +638,43 @@ class TestDecoratorWrapper(unittest.TestCase):
             self.assertTrue(isinstance(n, NameError))
             print(n)
 
+    def test_retry_until_exception_of_type_attempt_number(self):
+        try:
+            self.assertTrue(_retryable_test_with_unless_exception_type_name(
+                NameErrorUntilCount(5)))
+        except NameError as e:
+            s = _retryable_test_with_unless_exception_type_name.\
+                retry.statistics
+            self.assertTrue(s['attempt_number'] == 6)
+            print(e)
+        else:
+            self.fail("Expected NameError")
+
+    def test_retry_until_exception_of_type_no_type(self):
+        try:
+            # no input should catch all subclasses of Exception
+            self.assertTrue(
+                _retryable_test_with_unless_exception_type_no_input(
+                    NameErrorUntilCount(5))
+            )
+        except NameError as e:
+            s = _retryable_test_with_unless_exception_type_no_input.\
+                retry.statistics
+            self.assertTrue(s['attempt_number'] == 6)
+            print(e)
+        else:
+            self.fail("Expected NameError")
+
+    def test_retry_until_exception_of_type_wrong_exception(self):
+        try:
+            # two iterations with IOError, one that returns True
+            _retryable_test_with_unless_exception_type_name_attempt_limit(
+                IOErrorUntilCount(2))
+            self.fail("Expected RetryError")
+        except RetryError as e:
+            self.assertTrue(isinstance(e, RetryError))
+            print(e)
+
     def test_defaults(self):
         self.assertTrue(_retryable_default(NoNameErrorAfterCount(5)))
         self.assertTrue(_retryable_default_f(NoNameErrorAfterCount(5)))
